from typing import AsyncIterator, List

from aiostream import stream


async def merge_streams(streams: List[AsyncIterator]) -> AsyncIterator:
    if len(streams) > 1:
        merged_stream = stream.merge(streams[0], *streams[1:])
        async with merged_stream.stream() as s:
            async for value in s:
                yield value
    elif len(streams) == 1:
        async for v in streams[0]:
            yield v



import json
from functools import cache
from typing import Any, Optional

from httpx import Timeout, AsyncClient, HTTPStatusError, codes
from httpx_sse import EventSource
from pydantic_core import to_jsonable_python

from eidolon_ai_client.events import BaseStreamEvent
from eidolon_ai_client.util.logger import logger
from eidolon_ai_client.util.request_context import RequestContext


# noinspection PyShadowingNames
async def get_content(url: str, **kwargs):
    params = {"url": url, "headers": _headers()}
    async with AsyncClient(timeout=Timeout(5.0, read=600.0)) as client:
        response = await client.get(**params, **kwargs)
        await AgentError.check(response)
        return response.json()


async def post_content(url, json: Optional[Any] = None, **kwargs):
    params = {"url": url, "headers": _headers()}
    if json:
        params["json"] = to_jsonable_python(json)
    async with AsyncClient(timeout=Timeout(5.0, read=600.0)) as client:
        response = await client.post(**params, **kwargs)
        await AgentError.check(response)
        return response.json()


# noinspection PyShadowingNames
async def delete(url, **kwargs):
    params = {"url": url, "headers": _headers()}
    async with AsyncClient(timeout=Timeout(5.0, read=600.0)) as client:
        response = await client.delete(**params, **kwargs)
        await AgentError.check(response)
        return response.json()


@cache
def maybe_propagator():
    try:
        from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
        return TraceContextTextMapPropagator
    except ImportError:
        return None


# noinspection PyShadowingNames
def _headers():
    headers = RequestContext.headers
    TraceContextTextMapPropagator = maybe_propagator()
    if TraceContextTextMapPropagator:
        TraceContextTextMapPropagator().inject(carrier=headers)
    return headers


async def stream_content(url: str, body, **kwargs):
    body = to_jsonable_python(body)
    headers = _headers()
    headers["Accept"] = "text/event-stream"
    request = {"url": url, "json": body, "method": "POST", "headers": headers, **kwargs}
    async with AsyncClient(timeout=Timeout(5.0, read=600.0)) as client:
        async with client.stream(**request) as response:
            await AgentError.check(response)
            async for sse_event in EventSource(response).aiter_sse():
                if sse_event.data:
                    data = json.loads(sse_event.data)
                    event = BaseStreamEvent.from_dict(data)
                    yield event
                else:
                    logger.warning("Empty event from server")


class AgentError(Exception):
    message: str
    status_code: int
    response: Any

    def __init__(self, status_code: int, message: str, response=None):
        super().__init__(f"{status_code} ({codes.get_reason_phrase(status_code)}): {message}")
        self.message = message
        self.status_code = status_code
        self.response = response

    @classmethod
    async def check(cls, response, message_override=None):
        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            message = message_override or "".join([b async for b in e.response.aiter_text()])
            raise cls(e.response.status_code, message, response)



from __future__ import annotations

import os
from typing import List, Any, AsyncIterator, Optional
from urllib.parse import urljoin

import jsonref
from pydantic import BaseModel, Field, Extra

from eidolon_ai_client.events import StreamEvent, StartAgentCallEvent, AgentStateEvent
from eidolon_ai_client.util.aiohttp import stream_content, get_content, post_content, delete


def current_machine_url() -> str:
    return os.environ.get("EIDOLON_LOCAL_MACHINE", "http://localhost:8080")


class Machine(BaseModel):
    machine: str = Field(default_factory=current_machine_url)

    async def get_schema(self) -> dict:
        url = urljoin(self.machine, "openapi.json")
        json_ = await get_content(url)
        return jsonref.replace_refs(json_)

    def agent(self, agent_name: str) -> Agent:
        return Agent(machine=self.machine, agent=agent_name)

    async def processes(self) -> ProcessesResponse:
        url = urljoin(self.machine, "/processes")
        json_ = await get_content(url)
        print(json_)
        json_['processes'] = [{"machine":self.machine, **kwargs} for kwargs in json_['processes']]
        return ProcessesResponse(**json_)


class Agent(BaseModel):
    machine: str = Field(default_factory=current_machine_url)
    agent: str

    async def programs(self) -> List[str]:
        url = urljoin(self.machine, f"agents/{self.agent}/programs")
        return await get_content(url)

    async def create_process(self) -> ProcessStatus:
        url = urljoin(self.machine, "/processes")
        json_ = await post_content(url, json=dict(agent=self.agent))
        return ProcessStatus(**{"machine":self.machine, **json_})

    def process(self, process_id: str) -> Process:
        return Process(machine=self.machine, agent=self.agent, process_id=process_id)

    async def processes(self) -> ProcessesResponse:
        url = urljoin(self.machine, "/processes")
        json_ = await get_content(url)
        json_['processes'] = [{"machine":self.machine, **kwargs} for kwargs in json_['processes']]
        return ProcessesResponse(**json_)

    async def run_program(self, action_name: str, body: dict | BaseModel | str | None = None, **kwargs):
        process = await self.create_process()
        return await process.action(action_name, body, **kwargs)

    async def stream_program(self, action_name: str, body: Optional[Any] = None, **kwargs):
        process = await self.create_process()
        async for event in process.stream_action(action_name, body, **kwargs):
            yield event

    @classmethod
    def get(cls, location: str) -> Agent:
        """
        Convenience method to create Agents from dot notation. Ie: machine_loc.agent_name
        """
        if "." in location:
            parts = location.split(".")
            return cls(machine=".".join(parts[:-1]), agent=parts[-1])
        else:
            return cls(agent=location)


class Process(BaseModel):
    machine: str = Field(default_factory=current_machine_url)
    agent: str
    process_id: str

    async def action(self, action_name: str, body: dict | BaseModel | str | None = None, **kwargs) -> ProcessStatus:
        url = urljoin(self.machine, f"processes/{self.process_id}/agent/{self.agent}/actions/{action_name}")
        args = {
            "url": url, **kwargs
        }
        if body:
            args["json"] = body
        json_ = await post_content(**args)
        return ProcessStatus(**{"machine":self.machine, **json_})

    def stream_action(self, action_name: str, body: Optional[Any] = None, **kwargs) -> AgentResponseIterator:
        url = urljoin(self.machine, f"processes/{self.process_id}/agent/{self.agent}/actions/{action_name}")
        return AgentResponseIterator(stream_content(url, body, **kwargs))

    async def status(self) -> ProcessStatus:
        url = urljoin(self.machine, f"processes/{self.process_id}")
        json_ = await get_content(url)
        if 'machine' not in json_:
            json_['machine'] = self.machine
        return ProcessStatus(**json_)

    async def events(self):
        url = urljoin(self.machine, f"processes/{self.process_id}/events")
        return await get_content(url)

    async def delete(self) -> DeleteProcessResponse:
        deleted = await delete(urljoin(self.machine, f"processes/{self.process_id}"))
        return DeleteProcessResponse.model_validate(deleted)

    @classmethod
    def get(cls, stream_response: AgentResponseIterator):
        if not stream_response.machine or not stream_response.agent or not stream_response.process_id:
            raise ValueError("stream_response insufficiently iterated")
        return cls(machine=stream_response.machine, agent=stream_response.agent, process_id=stream_response.process_id)


class ProcessesResponse(BaseModel):
    total: int
    processes: List[ProcessStatus]
    next: Optional[str] = None


class ProcessStatus(Process, extra=Extra.allow):
    state: str
    available_actions: List[str]


class AgentResponseIterator(AsyncIterator[StreamEvent]):
    """
    This class is used to iterate over the responses from an agent call and store the state of the conversation after the stream is complete.

    For example::

        agent_it = agent.stream_program("program_name", "some data")
        async for event in agent_it:
            # ... do something with the event ...
        process_id = agent_it.process_id

    """

    data: AsyncIterator[StreamEvent]
    machine: str
    agent: str
    process_id: str
    state: str
    available_actions: List[str]

    def __init__(self, data: AsyncIterator[StreamEvent]):
        self.data = data.__aiter__()

    async def __anext__(self):
        try:
            event = await self.data.__anext__()
        except StopAsyncIteration:
            await self.iteration_complete()
            raise
        if event.is_root_and_type(StartAgentCallEvent):
            self.machine = event.machine
            self.agent = event.agent_name
            self.process_id = event.process_id
        elif event.is_root_and_type(AgentStateEvent):
            self.state = event.state
            self.available_actions = event.available_actions

        return event

    def __aiter__(self):
        return self

    async def iteration_complete(self):
        pass


class DeleteProcessResponse(BaseModel):
    process_id: str
    deleted: int



from __future__ import annotations

from contextvars import ContextVar
from typing import Any, Dict
from urllib.request import Request

from pydantic import BaseModel
from starlette.middleware.base import BaseHTTPMiddleware

from eidolon_ai_client.util.logger import logger

_request_context = ContextVar("request_context")


class _Record(BaseModel):
    key: str
    value: Any
    propagate: bool


def _get_context() -> Dict[str, _Record]:
    try:
        return _request_context.get()
    except LookupError:
        _request_context.set(dict())
        return _request_context.get()


class _RequestContextMeta(type):
    def __getitem__(self, key):
        return _get_context()[key].value

    def __repr__(self):
        return repr(_get_context())

    def __delitem__(self, key):
        del _get_context()[key]

    @staticmethod
    def set(key: str, value: str | Any, propagate=False):
        logger.debug(f"setting context {key}={value}, propagate={propagate}")
        if propagate and not isinstance(value, str):
            raise ValueError("can only propagate string values")
        if "," in key:
            raise ValueError("key cannot contain commas")
        _get_context()[key] = _Record(key=key, value=value, propagate=propagate)

    def get(self, key, default=...):
        context = _get_context()
        if default is ... and key not in context:
            raise KeyError(key)
        return self[key] if key in context else default

    @property
    def headers(self):
        to_propagate = {v.key: v.value for v in _get_context().values() if v.propagate}
        if to_propagate:
            to_propagate["X-Eidolon-Context"] = ",".join(f"{k}" for k in to_propagate.keys())
        return to_propagate


class RequestContext(metaclass=_RequestContextMeta):
    pass


class ContextMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        context_headers = request.headers.get("X-Eidolon-Context", "") or []
        if context_headers:
            context_headers = context_headers.split(",")
        for header in context_headers:
            try:
                RequestContext.set(header, request.headers[header], propagate=True)
            except KeyError:
                logger.warning(f"Expected context header {header} not found")

        return await call_next(request)



import logging
import re

logger = logging.getLogger("eidolon")


class CustomFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        arg_pattern = re.compile(r"%\((\w+)\)")
        arg_names = [x.group(1) for x in arg_pattern.finditer(self._fmt)]
        for field in arg_names:
            if field not in record.__dict__:
                record.__dict__[field] = None

        return super().format(record)



from __future__ import annotations

from abc import ABC
from enum import Enum
from pydantic import BaseModel, TypeAdapter
from typing import List, TypeVar, Generic, Any, AsyncIterator, Type, Literal, Dict, Optional


class Category(Enum):
    START = "start"
    INPUT = "input"
    END = "end"
    OUTPUT = "output"
    TRANSFORM = "transform"


T = TypeVar("T")


class BaseStreamEvent(BaseModel, ABC):
    stream_context: Optional[str] = None
    category: Category
    event_type: str

    def is_root_event(self):
        return self.stream_context is None

    def is_root_and_type(self, event_type: type):
        return self.stream_context is None and isinstance(self, event_type)

    def is_root_end_event(self):
        return self.is_root_and_type(EndStreamEvent)

    @classmethod
    def from_dict(cls, event_dict: Dict[str, Any]):
        # remove fields that are set automatically
        event_type = event_dict.pop("event_type")
        if "category" in event_dict:
            del event_dict["category"]
        if event_dict.get("stream_context", ...) is None:
            del event_dict["stream_context"]

        return _type_mapping[event_type](**event_dict)


class UserInputEvent(BaseStreamEvent):
    category: Literal[Category.INPUT] = Category.INPUT
    event_type: Literal["user_input"] = "user_input"
    input: Dict[str, Any]


class StartStreamContextEvent(BaseStreamEvent):
    category: Literal[Category.START] = Category.START
    event_type: Literal["context_start"] = "context_start"
    context_id: str
    title: str
    sub_title: str = ""

    def get_nested_context(self):
        context = self.stream_context + "." if self.stream_context else ""
        return context + self.context_id


class EndStreamContextEvent(BaseStreamEvent):
    category: Literal[Category.START] = Category.END
    event_type: Literal["context_end"] = "context_end"
    context_id: str


class ToolCallStartEvent(StartStreamContextEvent):
    event_type: Literal["tool_call_start"] = "tool_call_start"
    tool_call: ToolCall
    is_agent_call: bool = False


class StartAgentCallEvent(BaseStreamEvent):
    category: Literal[Category.START] = Category.START
    event_type: Literal["agent_call"] = "agent_call"
    machine: str
    agent_name: str
    call_name: str
    process_id: str


class OutputEvent(BaseStreamEvent, ABC):
    category: Literal[Category.OUTPUT] = Category.OUTPUT
    content: Any

    @staticmethod
    def get(content: T, **kwargs):
        if isinstance(content, str):
            return StringOutputEvent(content=content, **kwargs)
        else:
            return ObjectOutputEvent[T](content=content, **kwargs)


class LLMToolCallRequestEvent(BaseStreamEvent):
    category: Literal[Category.OUTPUT] = Category.OUTPUT
    event_type: Literal["llm_tool_call_request"] = "llm_tool_call_request"
    tool_call: ToolCall


class StringOutputEvent(OutputEvent):
    event_type: Literal["string"] = "string"
    content: str


class ObjectOutputEvent(OutputEvent, Generic[T]):
    event_type: Literal["object"] = "object"
    content: T


# note EndStreamEvent does not need to reference the type of event it ends since this is captured by context
class EndStreamEvent(BaseStreamEvent, ABC):
    category: Literal[Category.END] = Category.END
    event_type: Literal["error", "success", "canceled"]


class SuccessEvent(EndStreamEvent):
    event_type: Literal["success"] = "success"


class CanceledEvent(EndStreamEvent):
    event_type: Literal["canceled"] = "canceled"


class ErrorEvent(EndStreamEvent):
    event_type: Literal["error"] = "error"
    reason: str
    details: dict = {}


class AgentStateEvent(BaseStreamEvent):
    category: Literal[Category.TRANSFORM] = Category.TRANSFORM
    event_type: Literal["agent_state"] = "agent_state"
    state: str
    available_actions: List[str] = None  # this is filled in by the server, agents should leave the default


StreamEvent = (
    StartAgentCallEvent  # todo, this smells like UserInputEvent and StartAgentCallEvent
    | ToolCallStartEvent
    | StartStreamContextEvent
    | EndStreamContextEvent
    | LLMToolCallRequestEvent
    | StringOutputEvent
    | ObjectOutputEvent
    | SuccessEvent
    | CanceledEvent
    | ErrorEvent
    | AgentStateEvent
    | UserInputEvent
)

_type_mapping = {c.model_fields["event_type"].annotation.__args__[0]: c for c in StreamEvent.__args__}


async def convert_output_object(it: AsyncIterator[StreamEvent], output_format: Type[T]) -> AsyncIterator[StreamEvent]:
    model = TypeAdapter(output_format)
    async for event in it:
        if event.is_root_and_type(ObjectOutputEvent):
            event.content = model.validate_python(event.content)
        yield event


class ToolCall(BaseModel):
    tool_call_id: str
    name: str
    arguments: Dict[str, Any]



from typing import List, Any, Optional, Tuple

from pydantic import BaseModel

from eidolon_ai_client.client import Process, Agent, ProcessStatus, AgentResponseIterator, DeleteProcessResponse
from eidolon_ai_client.events import StartStreamContextEvent, EndStreamContextEvent
from eidolon_ai_client.util.stream_collector import merge_streams


class GroupConversation:
    def __init__(self, agents: List[ProcessStatus]):
        self.agents = agents

    @classmethod
    async def create(cls, agents: List[str]):
        agent_pids = []
        for agent in agents:
            agent_pids.append(await Agent.get(agent).create_process())
        return cls(agent_pids)

    @classmethod
    async def restore(cls, processes: List[ProcessStatus]):
        return cls(processes)

    async def action(self, action_name: str, body: dict | BaseModel | str | None = None, **kwargs) -> List[ProcessStatus]:
        statuses = []
        for agent in self.agents:
            statuses.append(await agent.action(action_name, body, **kwargs))
        self.agents = statuses
        return statuses

    async def stream_action(self, action_name: str, body: Optional[Any] = None, **kwargs) -> AgentResponseIterator:
        async def run_one(agent_name: str, agent_pid: str):
            yield StartStreamContextEvent(context_id=agent_name, title=agent_name)
            try:
                async for a_event in (
                        Agent.get(agent_name).process(agent_pid).stream_action(action_name, body, **kwargs)
                ):
                    a_event.stream_context = agent_name
                    yield a_event
            finally:
                yield EndStreamContextEvent(context_id=agent_name)

        tasks = []
        for process in self.agents:
            tasks.append(run_one(process.agent, process.process_id))

        combined_calls = merge_streams(tasks)
        async for event in combined_calls:
            yield event

        self.agents = await self.status()

    async def status(self) -> List[ProcessStatus]:
        statuses = []
        for agent in self.agents:
            statuses.append(await agent.status())
        return statuses

    async def delete(self) -> List[Tuple[Process, DeleteProcessResponse]]:
        deleted = []
        for agent in self.agents:
            deleted.append((agent, await agent.delete()))
        return deleted
